; 3x3 Lighted Pushbutton puzzle game
; Copyright 2009, Matt Rusnak. This file may be used for only for
;   non-commerical purposes.
; PIC16F505
; See schematic. Uses input columns for pushbuttons, with internal weak
; pull-ups. This version implements only the options for low and original
; difficulty. The options "menu" is fully functional, but other options have
; not yet been implemented, including the depth (number of bits each switch
; should change), fade effect, and demo mode.
; ----------------------------------------------------------------------------
;  Note: Switch input shared with ICSP -- pushbuttons WILL NOT WORK PROPERLY
;        while programmer is connected!
; ----------------------------------------------------------------------------
; Instructions:
; - When the circuit is turned on, it displays a moving pattern on the LEDs,
;   the "StartupSeq" state, waiting for the user to press any button. (This
;   helps randomize the game.
; - Press and release any button to start the game, or hold the center button
;   to go to the "options" menu. The only implemented options are switches 0
;   and 1, which work like radio buttons. Press switch 0 to choose the
;   "original" puzzle, or press switch 1 to swith to the regular random
;   puzzle (the default). Press switch 8 (with the blinking LED in the
;   lower-right corner) to exit the options menu.
; - Once the game is started, pressing any switch will toggle specific LEDs,
;   which only change upon reset or winning, in the random mode (default).
; - The object of the game is to turn off all of the LEDs -- the game can
;   always be won because the setup state starts with all LEDs off, and
;   simulates toggling random switches.
; - Upon winning, LEDs will show a flickering/rotating display for a few
;   seconds, and then return to the startup display. There are 4 startup
;   display patterns, which are chosen randomly.
; ----------------------------------------------------------------------------
; RC0-RC2 - columns for LEDs with current-limiting resistors
;    make high to light LEDs (one row at a time, because of shared resistors)
; RC3-RC5 - rows used for both LEDs and switches
;    make low when lighting LEDs or pulling down switches
; RB0, RB1, RB3 - ICSP shared with columns for normally-open switches
;    used as input with internal weak pull-ups when reading switches
; RB2 - unused, output low to limit wasted current
; RB4 - Win signal - make this high when user has won
; RB5 - Lose signal - make high when user has lost (if game has an end)
; (RB4 & RB5 Win/Lose signals not implemented in this version)
; Macros SetOptions_PullupOff and SetOptions_PullupOn are used to set the
;    whole option register. If other option register changes are needed,
;    this may have to change.
; TMR0 with 1:1 prescaler used for random puzzle setup
; ----------------------------------------------------------------------------
; This code uses the "errorlevel" command to suppress messages about page &
; bank changes temporarily, after code has been tested.
;   errorlevel -302 	; suppresses bank bits message for accessing variables
;   errorlevel -306		; suppresses page bit message for call/goto

	#include <p16f505.inc>
	__CONFIG	_MCLRE_OFF & _CP_OFF & _WDT_OFF & _IntRC_OSC_RB4EN

; -------------------- Constants --------------------

	; AppState values - state of the application
	constant StateRunning=0
	constant StateStartupSeqInit=1
	constant StateStartupSeq=2
	constant StateGameSetup=3
	constant StateWinSeq=4
	constant StateGameOptions=5
	; bit numbers in PORTB, for the switch columns
	constant SwitchCol0=0
	constant SwitchCol1=1
	constant SwitchCol2=3
	; bit numbers in PORTC for LED columns, and shared rows
	constant LEDCol0=0
	constant LEDCol1=1
	constant LEDCol2=2
	constant Row0=3
	constant Row1=4
	constant Row2=5
	; bits for GameOptions variable. normal game is low difficulty, other
	;   options off. difficulty is set up as masks to AND with GameOptions
	;   variable. options to toggle are listed as bit numbers.
	constant OptGameNormal=b'00001010'		; all options off, except low diff
											;   & low depth
	constant OptGameDiffMask=b'00000111'	; bits 0-2 for difficulty
	constant OptGameDiffOriginal=b'11111001'; masks, which bits to turn off	
	constant OptGameDiffLow=b'11111010'		;   difficulty is in bits 0-2.
	constant OptGameDiffHigh=b'11111100'
	constant OptGameDepthMask=b'00111000'	; bits 3-5 for depth (LEDs per sw)
	constant OptGameDepthLow=b'11001111'
	constant OptGameDepthMed=b'11010111'
	constant OptGameDepthHigh=b'11100111'
	constant OptGameFadeEffect=b'01000000'	; fade changed LEDs, or instant chg
	constant OptGameDemo=b'10000000'		; run demo of startup displays

; -------------------- Variables --------------------

; ---------- shared across all banks ----------
	cblock 0x08
Delay0		; delay counters for display timing
Delay1
Scratch0	; scratch registers, may not need all of these
Scratch1
Scratch2
Scratch3
Scratch4
Scratch5
	endc

; ---------- bank 0 ---------- 
	cblock 0x10
Display0	; one display variable for each LED, 8 bits for brightness levels
Display1
Display2
Display3
Display4
Display5
Display6
Display7
Display8
AppState		; state of main application (startup, running, win seq, etc.)
SwState0Prev	; states of swtiches, current & previous
SwState0Curr	;   SwState0x contains data for switches 0-7
SwState1Prev	;   SwState1x contains data for switch 8 (open for expansion
SwState1Curr	;   to 4x4 / 16 switches, and we have enough memory to spare)
GameOptions		; holds game options, few bits each, defined as constants above
OptionCounter	; counts how long a button is held, to enter options setup
	endc

; ---------- bank 1 ----------
	cblock 0x30
MasksL0		; first 8 bits of masks, one for each switch. e.g., user presses
MasksL1		;   switch 0, use MasksL0 and MasksH0 to determine which LEDs
MasksL2		;   to toggle.
MasksL3
MasksL4
MasksL5
MasksL6
MasksL7
MasksL8
	endc

; ---------- bank2 ----------
	cblock 0x50
MasksH0		; high bits of masks, see description for MasksL0
MasksH1
MasksH2
MasksH3
MasksH4
MasksH5
MasksH6
MasksH7
MasksH8
	endc

; ---------- bank 3 ---------- 
	cblock 0x70
Random0
Random1
Random2
Random3
Random4
Random5
Random6
Random7
Random8
Random9
	endc


; -------------------- Macros --------------------

; These two macros can be used to set the OPTION register,
; to control the weak pullups.
SetOptions_PullupOff macro
	movlw b'11011111'	; no wakeup, pullups off, no RC5 as timer
	option
  endm

SetOptions_PullupOn macro
	movlw b'10011111'	; no wakeup, pullups on, no RC5 as timer
	option
  endm

; -------------------- Main --------------------

	org 0
Init:
	clrf FSR				; this makes sure bank bits are 0.
	SetOptions_PullupOff	; turn off PORTB pull-ups (along with other option init)
	movlw b'00111000'		; PORTC columns low, rows high
	movwf PORTC
	movlw b'00000000'		; PORTC all all output
	tris PORTC
	movlw b'00000000'		; PORTB all low (most are going to be inputs anyway)
	movwf PORTB
	movlw b'00001011'		; RB0, RB1, RB3 inputs, rest output
	tris PORTB

	clrf Delay0				; clear the display timing variables
	clrf Delay1

	movlw OptGameNormal		; set to default options
	movwf GameOptions

	; make sure Random0 is 0 the first time through, so startup pattern 0
	;   is used on first run.
	movlw Random0
	movwf FSR
	errorlevel -302
	clrf Random0
	errorlevel +302
	clrf FSR

	movlw StateStartupSeqInit	; start off in StartupSeqInit, which sets vars
	movwf AppState				;   then changes to StartupSeq state

	movlw 0xFF				; set all switches to "not pressed" state
	movwf SwState0Prev
	movwf SwState0Curr
	movwf SwState1Prev
	movwf SwState1Curr

Main:
	; read switch states & debounce
	pagesel ReadSwitches
	errorlevel -306
	call ReadSwitches
	errorlevel +306
	pagesel Main


	; use AppState to choose which state to run
	movf AppState, w
	addwf PCL, f					; add AppState to PCL
	goto ProcessStateRunning		; 0
	goto ProcessStateStartupSeqInit	; 1
	goto ProcessStateStartupSeq		; 2
	goto ProcessStateGameSetup		; 3
	goto ProcessStateWinSeq			; 4
	goto ProcessStateGameOptions	; 5
	; code from state processing all uses "goto" to return here
StateComplete:
	; always call UpdateDisplay, no matter what state
	pagesel UpdateDisplay
	errorlevel -306
	call UpdateDisplay
	errorlevel +306
	pagesel Main

	; loop forever
	goto Main

; -------------------- State processing --------------------
; Each state only has one entry point, the equivalent of a switch()
; statement, above. Instead of using the stack for subroutines, use goto,
; and instead of returning, goto one location when finished.

; -------------------- ProcessStateRunning --------------------
ProcessStateRunning:
	; Scratch0 is used for processing switch states (which were pressed)
	; Scratch1 is a counter for looping to test each switch
	; Scratch2 holds bit masks copied from MasksLx or MasksHx
	; Scratch3 holds pointer to next Masks var, changing each loop iteration
	; Scratch4 is a counter for looping through sets of switches (#1 is 0-7
	;   and #2 is just 8)

	; set up Scratch0 with first group of switch states
	movf SwState0Curr, w
	xorwf SwState0Prev, w
	andwf SwState0Prev, w	; process only switches that were pressed
	movwf Scratch0

	; Scratch1 is a loop counter, process 8 switches in first group
	movlw 0x08
	movwf Scratch1

	; Scratch2 needs no initialization

	; Scratch3 is a pointer used for FSR (can't keep in FSR, because
	;   that changes the bank needed for display variables.)
	movlw MasksL0
	movwf Scratch3

	; Scratch4 is a loop counter
	movlw 0x02
	movwf Scratch4

	; process switches -- if switch has just been pressed, toggle LEDs
	;   related to that state in the Masks variables.
ProcessStateRunning_SwitchLoop:
	btfss Scratch0, 0
		goto ProcessStateRunning_NextSw
	movf Scratch3, w
	movwf FSR			; point at Masks variables in bank 1
	movf INDF, w
	movwf Scratch2
	clrf FSR			; back to bank 0
	btfsc Scratch2, 0
		comf Display0, f
	btfsc Scratch2, 1
		comf Display1, f
	btfsc Scratch2, 2
		comf Display2, f
	btfsc Scratch2, 3
		comf Display3, f
	btfsc Scratch2, 4
		comf Display4, f
	btfsc Scratch2, 5
		comf Display5, f
	btfsc Scratch2, 6
		comf Display6, f
	btfsc Scratch2, 7
		comf Display7, f
	; load the last set of masks (currently one bit in this 3x3 puzzle),
	;   without changing Scratch3, since that is incremented each time through
	;   this loop.
	movlw 0x20			; MasksLx is always 0x20 lower address from MasksHx
	addwf Scratch3, w
	movwf FSR			; point at MasksH0 variables in bank 2
	movf INDF, w
	movwf Scratch2
	movlw Display0
	movwf FSR			; back to bank 0
	movf Scratch2, w
	btfsc Scratch2, 0
		comf Display8, f
	; since a switch was pressed, call Randomize, to add more entropy.
	pagesel Randomize
	errorlevel -306
	call Randomize
	errorlevel +306
	pagesel Main
ProcessStateRunning_NextSw
	rrf Scratch0, f			; set up to check next switch state
	incf Scratch3, f		; point Scratch3 at next set of masks
	decfsz Scratch1, f		; update loop counter
		goto ProcessStateRunning_SwitchLoop

	; check second group of switch states
	movf SwState1Curr, w
	xorwf SwState1Prev, w
	andwf SwState1Prev, w	; process only switches that were pressed
	movwf Scratch0

	; reset counter, to test only 1 switch next time through the loop
	movlw 0x01
	movwf Scratch1

	; process second group of switch states (currently only 1 switch)
	decfsz Scratch4, f
		goto ProcessStateRunning_SwitchLoop

	; test for win, if all display variables are 0, Z flag will be set.
	movlw 0x00
	iorwf Display0, w
	iorwf Display1, w
	iorwf Display2, w
	iorwf Display3, w
	iorwf Display4, w
	iorwf Display5, w
	iorwf Display6, w
	iorwf Display7, w
	iorwf Display8, w

	; if Z flag is not set, we're done with this state's processing
	btfss STATUS, Z
		goto StateComplete

	; if Z was set, clear display timers & change to win sequence state
	clrf Delay0
	clrf Delay1
	clrf OptionCounter
	movlw StateWinSeq
	movwf AppState

	goto StateComplete

; -------------------- ProcessStateStartupSeqInit --------------------
ProcessStateStartupSeqInit:
	; set up the display, for the startup sequence
	; only runs once to set up variables, then changes to StateStartupSeq
	; Scratch0 holds pattern selection calculation

	; make sure OptionCounter is 0.
	clrf OptionCounter

	; change banks & copy Random0 to wreg
	movlw Random0
	movwf FSR
	errorlevel -302
	movf Random0, w
	errorlevel +302
	clrf FSR		; return to bank 0


	; use that random value to pick a startup pattern
	andlw 0x03	; use right 2 bits, make this higher if more patterns are added.
	addwf PCL, f
	goto StartupPattern0
	goto StartupPattern1
	goto StartupPattern2
	goto StartupPattern3

StartupPattern0:
	movlw b'00001111'
	movwf Display0
	movlw b'11110000'
	movwf Display1
	movlw b'00011111'
	movwf Display2
	movlw b'11000000'
	movwf Display3
	movlw b'00111111'
	movwf Display4
	movlw b'11100000'
	movwf Display5
	movlw b'10000000'
	movwf Display6
	movlw b'01111111'
	movwf Display7
	movlw b'11111111'
	movwf Display8
	goto StartupPatternSelected
StartupPattern1:
	movlw b'00000000'
	movwf Display0
	movwf Display3
	movwf Display6
	movlw b'00000011'
	movwf Display1
	movwf Display4
	movwf Display7
	movlw b'00001111'
	movwf Display2
	movwf Display5
	movwf Display8
	goto StartupPatternSelected
StartupPattern2:
	movlw b'00000001'
	movwf Display0
	movwf Display2
	movwf Display6
	movwf Display8
	movlw b'00000111'
	movwf Display1
	movwf Display3
	movwf Display5
	movwf Display7
	movlw b'00111111'
	movwf Display4
	goto StartupPatternSelected
StartupPattern3:
	movlw b'01111111'
	movwf Display0
	movwf Display1
	movwf Display2
	movwf Display3
	movlw b'00000111'
	movwf Display4
	movlw b'00000001'
	movwf Display5
	movwf Display6
	movwf Display7
	movwf Display8
StartupPatternSelected:
	movlw StateStartupSeq
	movwf AppState
	goto StateComplete

; -------------------- ProcessStateStartupSeq --------------------
ProcessStateStartupSeq:
	; rotates bits left, with carry flag set until full, then carry flag reset
	;   until nearly empty for fade-in/out effect. don't turn LEDs all the way
	;   off since they cut out abruptly. This state processes switches upon
	;   release, so the user can't hold a switch while turning on the device,
	;   to come up with the same puzzle every time.
	; Scratch0 used when checking switch states, and later as a loop counter

ProcessStateStartupSeq_Input:
	; if any switches have been released, change to GameSetup state
	clrf Scratch0
	movf SwState0Curr, w	; get a copy of current switch states
	; test for changed switches
	xorwf SwState0Prev, w	; xor so only changed bits are set
	andwf SwState0Curr, w	; AND with current switch states, to see which are released
	btfss STATUS, Z
		incf Scratch0, f
	movf SwState1Curr, w
	xorwf SwState1Prev, w
	btfss STATUS, Z
		incf Scratch0, f
	; if Scratch0 has increased 1 or 2, a switch has been pressed, if 0, no input
	movf Scratch0, w
	btfsc STATUS, Z
		goto ProcessStateStartupSeq_NoInput
	; if we have had input, change to the GameSetup state and return early
	movlw StateGameSetup
	movwf AppState
	goto StateComplete
ProcessStateStartupSeq_NoInput:
	; change the display only when Delay0 is 0 (fade timing looks right at
	;   this speed)
	movf Delay0, f
	btfss STATUS, Z
		goto StateComplete

	; point FSR at Display0
	movlw Display0
	movwf FSR

	; init loop counter for 9 display variables (loop ends at bottom if 0)
	movlw 0x09
	movwf Scratch0
ProcessStateStartupSeq_Loop:
	; if bit 7 is set, display var is all 1s, reset carry
	bcf STATUS, C
	btfss INDF, 6
		bsf STATUS, C
	; rotate display variable left & point FSR to next one
	rlf INDF, f
	incf FSR, f
	decfsz Scratch0, f
		goto ProcessStateStartupSeq_Loop

	; test for switch 4 (center) held, for game options. this is inside the
	;   display update loop, so that the counter doesn't fill up too quickly.
	btfss SwState0Curr, 4
		incf OptionCounter, f
	btfsc SwState0Curr, 4
		clrf OptionCounter
	movf OptionCounter, w
	xorlw 0x16			; length button was held (# of display update loops)
	btfss STATUS, Z				; exit before changing states, if button
		goto StateComplete		;   wasn't held long enough
	movlw StateGameOptions		; change to GameOptions state, if enough
	movwf AppState
	goto StateComplete

; -------------------- ProcessStateGameSetup --------------------
ProcessStateGameSetup:
	; Scratch0 is a general temporary variable.
	; Scratch1-5 are used to hold copies of the Random variables while
	;   processing, since they're in another bank, and we need access the
	;   banks containing the masks.

	; randomize again - this helps make the first game random, and mixes up
	;   the variables a little more for any additional games.
	pagesel Randomize
	errorlevel -306
	call Randomize
	errorlevel +306
	pagesel Main

	; load the first set of random variables. will load Random5-9 when needed.
	movlw Random0
	movwf FSR
	errorlevel -302
	movf Random0, w
	movwf Scratch1
	movf Random1, w
	movwf Scratch2
	movf Random2, w
	movwf Scratch3
	movf Random3, w
	movwf Scratch4
	movf Random4, w
	movwf Scratch5
	errorlevel +302
	clrf FSR

	; go to simplest setup if OptGameDiffOriginal bit is set
	btfsc GameOptions, 0
		goto ProcessStateGameSetup_Original
	; go to low-difficulty setup if OptGameDiffLow bit is set
	btfsc GameOptions, 1
		goto ProcessStateGameSetup_DiffLow
	; otherwise, assume it's high difficulty
ProcessStateGameSetup_DiffHigh:
	; FIX - need game setup state code for high difficulty, using low for now
	pagesel GameSetup_DiffLow
	errorlevel -306
	call GameSetup_DiffLow
	errorlevel +306
	pagesel Main
	goto ProcessStateGameSetup_Shuffle

ProcessStateGameSetup_DiffLow:
	; running out of memory in this page -- call setup function in 2nd page
	pagesel GameSetup_DiffLow
	errorlevel -306
	call GameSetup_DiffLow
	errorlevel +306
	pagesel Main
	goto ProcessStateGameSetup_Shuffle

ProcessStateGameSetup_Original:
	errorlevel -302
	; set low masks
	movlw MasksL0
	movwf FSR
	movlw b'00001011'
	movwf MasksL0
	movlw b'00010111'
	movwf MasksL1
	movlw b'00100110'
	movwf MasksL2
	movlw b'01011001'
	movwf MasksL3
	movlw b'10111010'
	movwf MasksL4
	movlw b'00110100'
	movwf MasksL5
	movlw b'11001000'
	movwf MasksL6
	movlw b'11010000'
	movwf MasksL7
	movlw b'10100000'
	movwf MasksL8
	; set high masks
	movlw MasksH0
	movwf FSR
	movlw b'00000000'
	movwf MasksL0
	movlw b'00000000'
	movwf MasksL1
	movlw b'00000000'
	movwf MasksL2
	movlw b'00000000'
	movwf MasksL3
	movlw b'00000000'
	movwf MasksL4
	movlw b'00000001'
	movwf MasksL5
	movlw b'00000000'
	movwf MasksL6
	movlw b'00000001'
	movwf MasksL7
	movlw b'00000001'
	movwf MasksL8
	errorlevel +302
	clrf FSR	; back to bank 0
	; and set all LEDs on
	movlw 0xFF
	movwf Display0
	movwf Display1
	movwf Display2
	movwf Display3
	movwf Display4
	movwf Display5
	movwf Display6
	movwf Display7
	movwf Display8
	; end of Original game setup, don't "shuffle" like the others.
	goto ProcessStateGameSetup_Done

ProcessStateGameSetup_Shuffle:
	; set all display variables off
	movlw 0x00
	movwf Display0
	movwf Display1
	movwf Display2
	movwf Display3
	movwf Display4
	movwf Display5
	movwf Display6
	movwf Display7
	movwf Display8

	; get a couple random variables
	movlw Random0
	movwf FSR
	errorlevel -302
	movf Random8, w
	movwf Scratch1
	movf Random9, w
	movwf Scratch2
	clrf FSR

	; randomly toggle some of the variables, based on the masks (simulating
	;   keypresses)
	; go through the loop 9 times
	movlw 0x09
	movwf Scratch0

	; load the carry flag with a random bit
	rlf Scratch2, f

ProcessStateGameSetup_ShuffLoop:
	; randomly skip
	btfss Scratch1, 0
		goto ProcessStateGameSetup_ShuffNext

	; load the display variables from the MaskL & MaskH variables
	movlw MasksL0
	movwf FSR
	movf Scratch0, w
	addwf FSR, f
	decf FSR, f
	movf INDF, w
	movwf Scratch3
	movlw 0x20		; move to the same variable in the next bank
	addwf FSR, f
	movf INDF, w
	movwf Scratch4
	clrf FSR

	; update display variables
	btfsc Scratch3, 0
		comf Display0, f
	btfsc Scratch3, 1
		comf Display1, f
	btfsc Scratch3, 2
		comf Display2, f
	btfsc Scratch3, 3
		comf Display3, f
	btfsc Scratch3, 4
		comf Display4, f
	btfsc Scratch3, 5
		comf Display5, f
	btfsc Scratch3, 6
		comf Display6, f
	btfsc Scratch3, 7
		comf Display7, f
	btfsc Scratch4, 0
		comf Display8, f

ProcessStateGameSetup_ShuffNext:
	; rotate to get another random bit
	rrf Scratch1, f

	; decrement counter & loop if needed
	decfsz Scratch0, f
		goto ProcessStateGameSetup_ShuffLoop

ProcessStateGameSetup_Done:
	; game setup is done, change to Running state
	movlw StateRunning
	movwf AppState
	goto StateComplete

; -------------------- ProcessStateWinSeq --------------------
ProcessStateWinSeq:
	; Shows an interesting display for a little while, then goes back to
	;   StateStartupSeqInit. Button presses are ignored during the win
	;   sequence.
	; OptionCounter is used to count how many times through the win sequence
	;   code, since the option menu can only be entered during the anyway.

	; call Randomize upon winning. we'll randomize again after the win sequence,
	;   when the user presses another button to exit the startup display.
	pagesel Randomize
	errorlevel -306
	call Randomize
	errorlevel +306
	pagesel Main

	; start with all LEDs off
	clrf Display0
	clrf Display1
	clrf Display2
	clrf Display3
	clrf Display4
	clrf Display5
	clrf Display6
	clrf Display7
	clrf Display8

	; alternately, just leave all LEDs off, to make them flicker
	btfss Delay0, 6
		goto StateComplete

	; make a spinning pattern
	btfsc Delay1, 1
		goto ProcessStateWinSeq_HighPattern
	btfss Delay0, 7
		goto ProcessStateWinSeq_Patt1
	goto ProcessStateWinSeq_Patt2
ProcessStateWinSeq_HighPattern:
	btfss Delay0, 7
		goto ProcessStateWinSeq_Patt3
	goto ProcessStateWinSeq_Patt4

ProcessStateWinSeq_Patt1:
	comf Display0, f
	comf Display8, f
	goto ProcessStateWinSeq_CheckEnd
ProcessStateWinSeq_Patt2:
	comf Display1, f
	comf Display7, f
	goto ProcessStateWinSeq_CheckEnd
ProcessStateWinSeq_Patt3:
	comf Display2, f
	comf Display6, f
	goto ProcessStateWinSeq_CheckEnd
ProcessStateWinSeq_Patt4:
	comf Display5, f
	comf Display3, f
	; no need for a "goto", since we're already here
ProcessStateWinSeq_CheckEnd:
	; end this state, when OptionCounter gets high enough
	movlw 0x40
	xorwf Delay1, w
	btfss STATUS, Z
		goto StateComplete

	; when finished, reset the OptionCounter, and change to StartupSeqInit
	clrf OptionCounter
	movlw StateStartupSeqInit
	movwf AppState

	goto StateComplete

; -------------------- ProcessStateGameOptions --------------------
ProcessStateGameOptions:
	; Scratch0 used as a counter.
	; Scratch1 used to hold processed switch states & GameOptions while
	;   processing display update.

	; switch 8 (last button) exits this state
	movf SwState1Curr, w
	xorwf SwState1Prev, w
	andwf SwState1Curr, w	; process only switches that were released
	btfsc STATUS, Z			; if button is pressed, continue with rest of state
		goto ProcessStateGameOptions_noExit
	movlw StateStartupSeqInit
	movwf AppState
	goto StateComplete
ProcessStateGameOptions_noExit:
	movf SwState0Curr, w
	xorwf SwState0Prev, w
	andwf SwState0Prev, w	; process only switches that were pressed
	movwf Scratch1

	; handle any switch presses
ProcessStateGameOptions_Opt0:
	btfss Scratch1, 0
		goto ProcessStateGameOptions_Opt1
	movlw OptGameDiffMask
	iorwf GameOptions, f
	movlw OptGameDiffOriginal
	andwf GameOptions, f
ProcessStateGameOptions_Opt1:
	btfss Scratch1, 1
		goto ProcessStateGameOptions_Opt2
	movlw OptGameDiffMask
	iorwf GameOptions, f
	movlw OptGameDiffLow
	andwf GameOptions, f
ProcessStateGameOptions_Opt2:
	btfss Scratch1, 2
		goto ProcessStateGameOptions_Opt3
	movlw OptGameDiffMask
	iorwf GameOptions, f
	movlw OptGameDiffHigh
	andwf GameOptions, f
ProcessStateGameOptions_Opt3:
	btfss Scratch1, 3
		goto ProcessStateGameOptions_Opt4
	movlw OptGameDepthMask
	iorwf GameOptions, f
	movlw OptGameDepthLow
	andwf GameOptions, f
ProcessStateGameOptions_Opt4:
	btfss Scratch1, 4
		goto ProcessStateGameOptions_Opt5
	movlw OptGameDepthMask
	iorwf GameOptions, f
	movlw OptGameDepthMed
	andwf GameOptions, f
ProcessStateGameOptions_Opt5:
	btfss Scratch1, 5
		goto ProcessStateGameOptions_Opt6
	movlw OptGameDepthMask
	iorwf GameOptions, f
	movlw OptGameDepthHigh
	andwf GameOptions, f
ProcessStateGameOptions_Opt6:
	btfss Scratch1, 6
		goto ProcessStateGameOptions_Opt7
	; toggle OptGameFadeEffect bit in GameOptions
	movf GameOptions, w
	andlw OptGameFadeEffect
	xorlw OptGameFadeEffect
	bcf GameOptions, 6		; clear fade effect bit, does not affect Z
	btfss STATUS, Z			; if Z, bit was clear originally, so set it
		bsf GameOptions, 6
ProcessStateGameOptions_Opt7:
	btfss Scratch1, 7
		goto ProcessStateGameOptions_Disp
	movf GameOptions, w
	andlw OptGameDemo
	xorlw OptGameDemo
	bcf GameOptions, 7		; clear demo bit, does not affect Z
	btfss STATUS, Z			; if Z, bit was clear originally, so set it
		bsf GameOptions, 7
	; FIX - Option 7 is to demo displays, using OptGameDemoBit, need to implement.

	; done processing options, go on to display updates.
ProcessStateGameOptions_Disp:

	; copy GameOptions to Scratch1
	movf GameOptions, w
	movwf Scratch1

	; clear display
	clrf Display0
	clrf Display1
	clrf Display2
	clrf Display3
	clrf Display4
	clrf Display5
	clrf Display6
	clrf Display7
	clrf Display8

	; FIX - do processing here, to make blinking/fading/etc. if desired

	; set display based on game options
	movlw 0xFF
	btfsc Scratch1, 0
		movwf Display0
	btfsc Scratch1, 1
		movwf Display1
	btfsc Scratch1, 2
		movwf Display2
	btfsc Scratch1, 3
		movwf Display3
	btfsc Scratch1, 4
		movwf Display4
	btfsc Scratch1, 5
		movwf Display5
	btfsc Scratch1, 6
		movwf Display6
	btfsc Scratch1, 7
		movwf Display7

	; make the last LED blink lightly (exit button)
	movwf b'0000101'
	btfss Delay1, 2
		movwf Display8

	goto StateComplete

; -------------------- Subroutines --------------------
	org 0x200
; subroutines start at 0x200, because PIC16F505's "call" instruction can only
; call addresses in the lower half of either page.

; ------------------------- Randomize -------------------------
Randomize:
	; takes Timer0, and combines it with values already in the Random0-3 variables.
	;   this is called after switch presses, so Timer0 is not predictable.
	movf TMR0, w

	; change to bank3, with Random0-3 variables.
	bsf FSR, 6
	bsf FSR, 5

	; FIX -- it would be better if these didn't depend on each other so much
	;   but this is close enough to random for our purposes.
	; mix up the random variables.
 	errorlevel -302		; turn off bank bits warning temporarily.
	movwf Random0
	xorwf Random3, w
	movwf Random5
	xorwf Random4, w
	movwf Random7
	xorwf Random6, w
	xorwf Random4, f
	xorwf Random2, w
	xorwf Random3, f
	xorwf Random7, f
	xorwf Random1, w
	rlf Random2, f
	rlf Random2, f
	xorwf Random2, f
	xorwf Random2, w
	movwf Random1
	xorwf Random5, f
	xorwf Random3, w
	xorwf Random8, f
	movwf Random9
	errorlevel +302		; turn bank bits warning back on.

 	; change back to bank0
	clrf FSR
	
	retlw 0

; ------------------------- UpdateDisplay -------------------------
UpdateDisplay:
	; Scratch0 used to calculate which bits to use
	; Scratch1 used to hold a bit mask for use with lighting the LEDs

	; update display timing counters
	incf Delay0, f
	btfsc STATUS, Z
		incf Delay1, f

; FIX - can save processing time and program memory, if we use 1 byte var
; and simply clear carry flag, rlf, and reset to 0x01 if carry is set.
; might also want to rotate only once every 2-4 times through the loop,
; for more noticeable variation.
	; choose which bit of Display variables to use for LED states, for
	;   brightness control based on lowest 3 bits of display timing counters.
	movf Delay0, w
	andlw b'00000111'
	movwf Scratch0
	movlw b'00000001'		; bit 0 by default
	movwf Scratch1
	bcf STATUS, C			; no carry in the coming rlf's
	btfss Scratch0, 2
		goto UpdateDisplay_SkipBit2
	movlw b'00010000'		; if bit 2 was set in counter, result is at
	movwf Scratch1			;   least bit 4 for the mask
UpdateDisplay_SkipBit2:
	btfsc Scratch0, 1		; shift left if bit 1 (1 of 2) & avoid goto
		rlf Scratch1, f
	btfsc Scratch0, 1		; shift left if bit 1 (2 of 2) & avoid goto
		rlf Scratch1, f
	btfsc Scratch0, 0		; and last, if bit 0 is set, rotate left once more
		rlf Scratch1, f

	; set up PORTC for lighting LEDs
	movlw b'00111000'	; PORTC columns low, rows high
	movwf PORTC
	movlw b'00000000'	; make all outputs
	tris PORTC

; FIX - can this be done more efficiently?
	; ---------- update display - Row0 ----------
	bcf PORTC, Row0
	movf Scratch1, w
	andwf Display0, w
	btfss STATUS, Z
		bsf PORTC, LEDCol0
	movf Scratch1, w
	andwf Display1, w
	btfss STATUS, Z
		bsf PORTC, LEDCol1
	movf Scratch1, w
	andwf Display2, w
	btfss STATUS, Z
		bsf PORTC, LEDCol2
	; LEDs off - one at a time, with delays so brightness is even
	bcf PORTC, LEDCol0
	goto $+1
	nop
	bcf PORTC, LEDCol1
	goto $+1
	nop
	bcf PORTC, LEDCol2
	; done with this row
	bsf PORTC, Row0
	; ---------- update display - Row1 ----------
	bcf PORTC, Row1
	movf Scratch1, w
	andwf Display3, w
	btfss STATUS, Z
		bsf PORTC, LEDCol0
	movf Scratch1, w
	andwf Display4, w
	btfss STATUS, Z
		bsf PORTC, LEDCol1
	movf Scratch1, w
	andwf Display5, w
	btfss STATUS, Z
		bsf PORTC, LEDCol2
	; LEDs off - one at a time, with delays so brightness is even
	bcf PORTC, LEDCol0
	goto $+1
	nop
	bcf PORTC, LEDCol1
	goto $+1
	nop
	bcf PORTC, LEDCol2
	; done with this row
	bsf PORTC, Row1
	; ---------- update display - Row2 ----------
	bcf PORTC, Row2
	movf Scratch1, w
	andwf Display6, w
	btfss STATUS, Z
		bsf PORTC, LEDCol0
	movf Scratch1, w
	andwf Display7, w
	btfss STATUS, Z
		bsf PORTC, LEDCol1
	movf Scratch1, w
	andwf Display8, w
	btfss STATUS, Z
		bsf PORTC, LEDCol2
	; LEDs off - one at a time, with delays so brightness is even
	bcf PORTC, LEDCol0
	goto $+1
	nop
	bcf PORTC, LEDCol1
	goto $+1
	nop
	bcf PORTC, LEDCol2
	; done with this row
	bsf PORTC, Row2

	retlw 0

; ------------------------- ReadSwitches -------------------------
ReadSwitches:
	; reads switch states, stores 4 successive reads in scratch registers,
	;    and debounces that input at the end.
	; Scratch0 used as counter, to read switches 4x.
	;   after loop, Scratch0 is a temporary register for debouncing
	; Scratch1-Scratch5 used to hold switch input, each switch uses half a byte
	;	in order from 0 to 8. 4 set/reset in a row = state, otherwise ignore
	movlw 0x04
	movwf Scratch0
	clrf Scratch1
	clrf Scratch2
	clrf Scratch3
	clrf Scratch4
	clrf Scratch5

	; set up PORTC for switch column reads
	movlw b'00111000'	; PORTC columns low, rows high
	movwf PORTC
	movlw b'00000111'	; change LED columns to inputs (PORTC has no pull-ups)
	tris PORTC
	; turn on pull-ups for switch reads
	SetOptions_PullupOn
ReadSwitchesLoop:
	rlf Scratch1, f		; rotate all to the left each pass
	rlf Scratch2, f		;   wastes time on first pass, but better than wasting
	rlf Scratch3, f		;   space with code to skip it, and can't do it at the
	rlf Scratch4, f		;   end because it would shift off the first data read.
	rlf Scratch5, f
	; ---------- read switch states - Row0 ----------
	bcf PORTC, Row0
	btfsc PORTB, SwitchCol0
		bsf Scratch1, 4		; high nibble, switch 0
	btfsc PORTB, SwitchCol1
		bsf Scratch1, 0		; low nibble, switch 1
	btfsc PORTB, SwitchCol2
		bsf Scratch2, 4		; high nibble, switch 2
	bsf PORTC, Row0
	; ---------- read switch states - Row1 ----------
	bcf PORTC, Row1
	btfsc PORTB, SwitchCol0
		bsf Scratch2, 0		; low nibble, switch 3
	btfsc PORTB, SwitchCol1
		bsf Scratch3, 4		; high nibble, switch 4
	btfsc PORTB, SwitchCol2
		bsf Scratch3, 0		; low nibble, switch 5
	bsf PORTC, Row1
	; ---------- read switch states - Row2 ----------
	bcf PORTC, Row2
	btfsc PORTB, SwitchCol0
		bsf Scratch4, 4		; high nibble, switch 6
	btfsc PORTB, SwitchCol1
		bsf Scratch4, 0		; low nibble, switch 7
	btfsc PORTB, SwitchCol2
		bsf Scratch5, 4		; high nibble, switch 8
	bsf PORTC, Row2

	; loop until switches are read 4 times
	decfsz Scratch0, f
		goto ReadSwitchesLoop

	; move last good switch states to prev variables
	movf SwState0Curr, w
	movwf SwState0Prev
	movf SwState1Curr, w
	movwf SwState1Prev

	; debounce input from the scratch registers, and save in SwState registers.
	;   only change the bit if we have all 0's or all 1's.
	; ----- switch 0 -----
	movlw 0xF0
	andwf Scratch1, w
	btfsc STATUS, Z
		bcf SwState0Curr, 0		; if this nibble is all 0's, make it 0
	xorlw 0xF0
	btfsc STATUS, Z
		bsf SwState0Curr, 0		; if this nibble is all 1's, make it 1
	; ----- switch 1 -----
	movlw 0x0F
	andwf Scratch1, w
	btfsc STATUS, Z
		bcf SwState0Curr, 1		; if this nibble is all 0's, make it 0
	xorlw 0x0F
	btfsc STATUS, Z
		bsf SwState0Curr, 1		; if this nibble is all 1's, make it 1
	; ----- switch 2 -----
	movlw 0xF0
	andwf Scratch2, w
	btfsc STATUS, Z
		bcf SwState0Curr, 2		; if this nibble is all 0's, make it 0
	xorlw 0xF0
	btfsc STATUS, Z
		bsf SwState0Curr, 2		; if this nibble is all 1's, make it 1
	; ----- switch 3 -----
	movlw 0x0F
	andwf Scratch2, w
	btfsc STATUS, Z
		bcf SwState0Curr, 3		; if this nibble is all 0's, make it 0
	xorlw 0x0F
	btfsc STATUS, Z
		bsf SwState0Curr, 3		; if this nibble is all 1's, make it 1
	; ----- switch 4 -----
	movlw 0xF0
	andwf Scratch3, w
	btfsc STATUS, Z
		bcf SwState0Curr, 4		; if this nibble is all 0's, make it 0
	xorlw 0xF0
	btfsc STATUS, Z
		bsf SwState0Curr, 4		; if this nibble is all 1's, make it 1
	; ----- switch 5 -----
	movlw 0x0F
	andwf Scratch3, w
	btfsc STATUS, Z
		bcf SwState0Curr, 5		; if this nibble is all 0's, make it 0
	xorlw 0x0F
	btfsc STATUS, Z
		bsf SwState0Curr, 5		; if this nibble is all 1's, make it 1
	; ----- switch 6 -----
	movlw 0xF0
	andwf Scratch4, w
	btfsc STATUS, Z
		bcf SwState0Curr, 6		; if this nibble is all 0's, make it 0
	xorlw 0xF0
	btfsc STATUS, Z
		bsf SwState0Curr, 6		; if this nibble is all 1's, make it 1
	; ----- switch 7 -----
	movlw 0x0F
	andwf Scratch4, w
	btfsc STATUS, Z
		bcf SwState0Curr, 7		; if this nibble is all 0's, make it 0
	xorlw 0x0F
	btfsc STATUS, Z
		bsf SwState0Curr, 7		; if this nibble is all 1's, make it 1
	; ----- switch 8 -----
	movlw 0xF0
	andwf Scratch5, w
	btfsc STATUS, Z
		bcf SwState1Curr, 0		; if this nibble is all 0's, make it 0
	xorlw 0xF0
	btfsc STATUS, Z
		bsf SwState1Curr, 0		; if this nibble is all 1's, make it 1

	; turn pullups back off, so they do not affect LEDs
	SetOptions_PullupOff

	retlw 0

; ------------------------- GameSetup_DiffLow -------------------------
GameSetup_DiffLow:
	; Sets up the low-difficulty game by using the random values
	; placed in the Scratch variables before this function was called.
	; For each switch, the mask variables are set or cleared using the 
	; random bits, only for the LEDs near the switch, horizontally,
	; vertically, or diagonally -- except that the center switch does
	; not toggle diagonals.

	; lots of bank switching ahead, turn off bank bit warnings
	errorlevel -302

 	; set masks for switch 0
	movlw MasksL0
	movwf FSR
	clrf INDF
	bsf INDF, 0			; set this switch's LED
	btfsc Scratch1, 0
		bsf INDF, 1
	btfsc Scratch1, 1
		bsf INDF, 3
	btfsc Scratch1, 2
		bsf INDF, 4
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0

	; set masks for switch 1
	incf FSR, f
	clrf INDF
	bsf INDF, 1			; set this switch's LED
 	btfsc Scratch1, 3
		bsf INDF, 0
	btfsc Scratch1, 4
		bsf INDF, 2
	btfsc Scratch1, 5
		bsf INDF, 3
	btfsc Scratch1, 6
		bsf INDF, 4
	btfsc Scratch1, 7
		bsf INDF, 5
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0

	; set masks for switch 2
	incf FSR, f
	clrf INDF
	bsf INDF, 2			; set this switch's LED
	btfsc Scratch2, 0
		bsf INDF, 1
	btfsc Scratch2, 1
		bsf INDF, 4
	btfsc Scratch2, 2
		bsf INDF, 5
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0


	; set masks for switch 3
	incf FSR, f
	clrf INDF
	bsf INDF, 3			; set this switch's LED
	btfsc Scratch2, 3
		bsf INDF, 0
	btfsc Scratch2, 4
		bsf INDF, 1
	btfsc Scratch2, 5
		bsf INDF, 4
	btfsc Scratch2, 6
		bsf INDF, 6
	btfsc Scratch2, 7
		bsf INDF, 7
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0


	; set masks for switch 4 - this does not include diagonals (4 corners)
 	incf FSR, f
	clrf INDF
	bsf INDF, 4			; set this switch's LED
	btfsc Scratch3, 0
		bsf INDF, 1
	btfsc Scratch3, 1
		bsf INDF, 3
	btfsc Scratch3, 2
		bsf INDF, 5
	btfsc Scratch3, 3
		bsf INDF, 7
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0

	; set masks for switch 5
	incf FSR, f
	clrf INDF
	bsf INDF, 5			; set this switch's LED
	btfsc Scratch3, 3	; FIX - this random bit is used twice, incl for sw 4
		bsf INDF, 1
	btfsc Scratch3, 4
		bsf INDF, 2
	btfsc Scratch3, 5
		bsf INDF, 4
	btfsc Scratch3, 6
		bsf INDF, 7
	banksel MasksH0
	btfsc Scratch3, 7
		bsf INDF, 0
	banksel MasksL0

	; set masks for switch 6
	incf FSR, f
	clrf INDF
	bsf INDF, 6			; set this switch's LED
	btfsc Scratch4, 0
		bsf INDF, 3
	btfsc Scratch4, 1
		bsf INDF, 4
	btfsc Scratch4, 2
		bsf INDF, 7
	banksel MasksH0
	bcf INDF, 0
	banksel MasksL0

	; set masks for switch 7
	incf FSR, f
	clrf INDF
	bsf INDF, 7			; set this switch's LED
	btfsc Scratch5, 3
		bsf INDF, 3
	btfsc Scratch5, 4
		bsf INDF, 4
	btfsc Scratch5, 5
		bsf INDF, 5
	btfsc Scratch5, 6
		bsf INDF, 6
	banksel MasksH0
	btfsc Scratch5, 7
		bsf INDF, 0
	banksel MasksL0

	; FIX - should load another random variable - this is less random than it
	;   could be, but not easily predictable during play.

	; set masks for switch 8
	incf FSR, f
	clrf INDF
	btfsc Scratch4, 0
		bsf INDF, 4
	btfsc Scratch4, 1
		bsf INDF, 5
	btfsc Scratch4, 2
		bsf INDF, 7
	banksel MasksH0
	bsf INDF, 0			; set this switch's LED
	banksel MasksL0


	; change back to bank 0
	errorlevel +302
	clrf FSR

	retlw 0


	end